/*
 * This file is part of W3Olista, the statistics gathering program for
 * the WWW.                  This sourcecode is (c) Frank Pilhofer '95
 *
 * It is  distributed without warranty. The complete license terms for
 * this software  are included in  the file 'License.html'  in the doc
 * subdirectory. If  your copy came  without this file, you should get
 * it on the Web at    http://www.uni-frankfurt.de/~fp/Tools/W3Olista/
 * or contact the author,               fp@informatik.uni-frankfurt.de
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "toolbox.h"
#include "struct.h"
#include "getcgi.h"
#include "cgiproc.h"

/*
 * Process CGI data
 */

char **inhost=NULL, **exhost=NULL;
char **infile=NULL, **exfile=NULL;

/*
 * cgiencode especially has to exchange slashes in filenames to
 * something else ($)
 */

char *cgiencode (char *string)
{
  static char temp[1024];
  char *p1=string, *p2=temp;

  while (*p1) {
    if (isalpha (*p1) || isdigit (*p1))
      *p2++ = *p1++;
    else if (*p1 == '/') { /* special handling */
      *p2++ = '$';
      p1++;
    }
    else {
      sprintf (p2, "%%%02X", *p1);
      p1++; p2 += 3;
    }
  }
  *p2 = '\0';
  return temp;
}

char *cgidecode (char *string)
{
  static char temp[1024];
  char *p1=string, *p2=temp, c;

  while (*p1) {
    if (*p1 == '$') {
      *p2++ = '/';
      p1++;
    }
    else if (*p1 == '%' && isxdigit (*(p1+1)) && isxdigit (*(p1+2))) {
      c = tolower (*++p1);
      if (isdigit(c))       *p2  = (c-'0')    << 4;
      else if (isalpha(c))  *p2  = (c-'a'+10) << 4;
      else *p2 = ' ';
      c = tolower (*++p1);
      if (isdigit(c))       *p2 += c-'0';
      else if (isalpha(c))  *p2 += c-'a'+10;
      else *p2 = ' ';
      p1++; p2++;
    }
    else
      *p2++ = *p1++;
  }
  *p2 = '\0';
  return temp;
}

/*
 * Extract what should be included and excluded because of
 * CGI input data. Don't change data that's already there
 */

static char **AddIncludeExclude (char *prefix, char **old)
{
  char string[20], **new, **pptr, *ptr, *p2, c;
  int count=0, index=0, copy=0;

  if ((pptr=old) != NULL)
    while (*pptr++)
      count++;

  while (1) {
    sprintf (string, "%s%03d", prefix, ++index);

    if ((ptr = GetInputValue (string)) != NULL) {
      while (ptr) {
	p2 = strpbrk (ptr, ";, ");
	if (p2) { c = *p2; *p2 = '\0'; }
	if (*ptr) count++;
	if (p2) { *p2 = c; while (*p2==';'||*p2==','||*p2==' ') p2++; }
	ptr = p2;
      }
    }
    else
      break;
  }

  if (count == 0)
    return old;

  if ((new = (char **) malloc ((count + 1) * sizeof (char *))) == NULL)
    return NULL;

  index=copy=0;

  if (old) {
    for (; old[copy] && copy<count; copy++)
      new[copy] = old[copy];
  }

  while (copy<count) {
    sprintf (string, "%s%03d", prefix, ++index);
    
    if ((ptr=GetInputValue (string)) != NULL) {
      while (ptr) {
	p2 = strpbrk (ptr, ";, ");
	if (p2) { c = *p2; *p2 = '\0'; }
	if (*ptr) new[copy++] = _FP_strdup (cgidecode (ptr));
	if (p2) { *p2 = c; while (*p2==';'||*p2==','||*p2==' ') p2++; }
	ptr = p2;
      }
    }
    else
      break;
  }
  new[copy] = NULL;

  free (old);
  return new;
}

void GetIncludeExclude (void)
{
  inhost = AddIncludeExclude ("inhost", inhost);
  exhost = AddIncludeExclude ("exhost", exhost);
  infile = AddIncludeExclude ("infile", infile);
  exfile = AddIncludeExclude ("exfile", exfile);
}

void DelIncludeExclude (void)
{
  char **pptr;

  if ((pptr = inhost) != NULL) {
    while (*pptr != NULL)
      free (*pptr++);
    free (inhost);
    inhost = NULL;
  }
  if ((pptr = exhost) != NULL) {
    while (*pptr != NULL)
      free (*pptr++);
    free (exhost);
    exhost = NULL;
  }
  if ((pptr = infile) != NULL) {
    while (*pptr != NULL)
      free (*pptr++);
    free (infile);
    infile = NULL;
  }
  if ((pptr = exfile) != NULL) {
    while (*pptr != NULL)
      free (*pptr++);
    free (exfile);
    exfile = NULL;
  }
}

int GetCGIReport (int num, char *report)
{
  char string[20], *p1, *p2;
  int result=NOTYPE;

  sprintf (string, "rep%03d", num);

  if ((p1 = GetInputValue (string)) == NULL)
    return NOTYPE;
  if ((p2 = strchr (p1, ';')) == NULL)
    return NOTYPE;

  strcpy (report, cgidecode (p2+1));

  if (strncmp (p1, "host", 4) == 0)
    result = HOSTTYPE;
  else if (strncmp (p1, "file", 4) == 0)
    result = FILETYPE;
  else if (strncmp (p1, "date", 4) == 0)
    result = DATETYPE;

  return result;
}
