/*
 * getcgi.c
 *
 * Distributed by the GNU General Public License. Use and be happy.
 * Read http://www.uni-frankfurt.de/~fp/Tools/Getcgi.html for more
 * information.                     fp@informatik.uni-frankfurt.de
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include "toolbox.h"
#include "getcgi.h"

typedef struct _ifield {
  char *name;
  char *value;
  struct _ifield *NEXT;
} ifield;

static ifield *inputs  = NULL;
static int     errcode = 0;

char *readfile (FILE *file, int length)
{
  char *result;
  int size;

  if ((result = malloc (length + 1)) == NULL) {
    errcode = _CGI_NOMEM;
    return NULL;
  }

  if ((size = fread (result, sizeof (char), length, file)) < length)
    errcode = _CGI_ERRINP;
    
  result[size] = '\0';

  return result;
}

char *LoadInput (void)
{
  char *result, *method, *p;
  int len;

  if ((method = getenv ("REQUEST_METHOD")) == NULL)
    return NULL;

  if (strcmp (method, "GET") == 0) {
    if ((p = getenv ("QUERY_STRING")) == NULL)
      return NULL;
    else
      result = _FP_strdup (p);
  }
  else if (strcmp (method, "POST") == 0) {
    if ((len = atoi (getenv ("CONTENT_LENGTH"))) == 0)
      return NULL;

    result = readfile (stdin, len);
  }
  else
    return NULL;

  return result;
}

char *ParseString (char *instring)
{
  char *ptr1=instring, *ptr2=instring;

  if (instring == NULL)
    return instring;

  while (isspace (*ptr1))
    ptr1++;

  while (*ptr1) {
    if (*ptr1 == '+') {
      ptr1++; *ptr2++=' ';
    }
    else if (*ptr1 == '%' && isxdigit (*(ptr1+1)) && isxdigit (*(ptr1+2))) {
      ptr1++;
      *ptr2    = ((*ptr1>='0'&&*ptr1<='9')?(*ptr1-'0'):((char)toupper(*ptr1)-'A'+10)) << 4;
      ptr1++;
      *ptr2++ |= ((*ptr1>='0'&&*ptr1<='9')?(*ptr1-'0'):((char)toupper(*ptr1)-'A'+10));
      ptr1++;
    }
    else 
      *ptr2++ = *ptr1++;
  }
  while (ptr2>instring && isspace(*(ptr2-1)))
    ptr2--;

  *ptr2 = '\0';

  return instring;
}

int BreakString (char *input, char **left, char **right)
{
  static char *empty="";
  char *ptr;

  if (input == NULL) {
    *left = *right = NULL;
    return 0;
  }
  
  if ((ptr = strchr (input, '=')) == NULL) {
    *left  = _FP_strdup (input);
    *right = _FP_strdup (empty);
  }
  else {
    *ptr = '\0';
    *left  = _FP_strdup (input);
    *right = _FP_strdup (ptr+1);
    *ptr = '=';
  }
  return 1;
}

ifield *GetNextInput (char *input)
{
  ifield *result;

  if (input == NULL || *input == '\0')
    return NULL;

  if ((result = malloc (sizeof (ifield))) == NULL) {
    errcode = _CGI_NOMEM;
    return NULL;
  }

  memset (result, '\0', sizeof (ifield));

  if (BreakString (input, &(result->name), &(result->value)) == 0) {
    free (result->name);
    free (result->value);
    free (result);
    return NULL;
  }
  ParseString (result->name);
  ParseString (result->value);

  return result;
}

int AddVariables (char *p1, int delimit)
{
  ifield *iter = inputs;
  char *p2;

  if (p1 == NULL)
    return errcode;

  if (iter != NULL)
    while (iter->NEXT != NULL)
      iter = iter->NEXT;

  while (p1 != NULL) {
    if ((p2 = strchr (p1, delimit)) != NULL) {
      if (p1 == p2) {
	p1++;
	continue;
      }
      *p2 = '\0';
    }

    if (iter == NULL) {
      if ((inputs = iter = GetNextInput (p1)) == NULL)
	return errcode;
    }
    else {
      if ((iter->NEXT = GetNextInput (p1)) == NULL)
	break;

      if (iter->NEXT)
	iter = iter->NEXT;
    }
    if (p2) {
      *p2 = delimit;
      p1  = p2 + 1;
    }
    else
      break;
  }
  return errcode;
}

int InitInputValues (char *conffile, int argc, char *argv[])
{
  char *string, *p1, temp[256];
  FILE *datei;
  int index;

  errcode = _CGI_OK;

  if (conffile && (datei = fopen (conffile, "r"))) {
    while (!feof (datei)) {
      if ((p1 = fgets (temp, 255, datei)) == NULL)
	break;
      temp[255] = '\0';
      while (isspace (*p1))
	p1++;
      AddVariables (p1, '&');
    }
  }
  if ((string = LoadInput()) == NULL)
    if (errcode != _CGI_OK)
      return errcode;
  
  AddVariables (string, '&');
  free (string);

  if ((string = getenv ("PATH_INFO")) != NULL) {
    if ((p1 = _FP_strdup (string)) != NULL) { /* always play safe */
      AddVariables (p1, '/');
      free (p1);
    }
  }
  if (getenv ("REMOTE_HOST") && getenv ("REMOTE_ADDR")) {
    AddVariable ("Remote",   getenv ("REMOTE_HOST"));
    AddVariable ("RemoteIP", getenv ("REMOTE_ADDR"));
  }
  else if (getenv ("REMOTE_ADDR")) {
    AddVariable ("Remote",   getenv ("REMOTE_ADDR"));
    AddVariable ("RemoteIP", getenv ("REMOTE_ADDR"));
  }
  else {
    AddVariable ("Remote",   "Unknown");
    AddVariable ("RemoteIP", "Unknown");
  }
  if (getenv ("SCRIPT_NAME"))
    AddVariable ("SCRIPT_NAME", getenv ("SCRIPT_NAME"));
  if (getenv ("SERVER_NAME"))
    AddVariable ("SERVER_NAME", getenv ("SERVER_NAME"));
  if (getenv ("SERVER_PORT"))
    AddVariable ("SERVER_PORT", getenv ("SERVER_PORT"));

  for (index=1; index<argc; index++)
    AddVariables (argv[index], '&');

  return errcode;
}

char *GetInputValue (char *name)
{
  ifield *iter = inputs;

  while (iter != NULL) {
    if (_FP_stricmp (name, iter->name) == 0)
      break;
    else
      iter = iter->NEXT;
  }

  return (iter) ? iter->value : NULL;
}

char *GetInputNameByNum (int number)
{
  ifield *iter = inputs;

  while (iter && number) {
    iter = iter->NEXT;
    number--;
  }
  return (iter) ? iter->name : NULL;
}

char *GetInputValueByNum (int number)
{
  ifield *iter = inputs;

  while (iter && number) {
    iter = iter->NEXT;
    number--;
  }
  return (iter) ? iter->value : NULL;
}

void AddVariable (char *name, char *value)
{
  ifield *iter = inputs, *new;
  static char *ersatz = "";

  if (value == NULL)
    value = ersatz;

  if ((new = (ifield *) malloc (sizeof (ifield))) == NULL) {
    errcode = _CGI_NOMEM;
    return;
  }
  memset (new, '\0', sizeof (ifield));

  if (iter == NULL) {
    inputs     = new;
    new->name  = _FP_strdup (name);
    new->value = _FP_strdup (value);
  }
  else {
    while (iter->NEXT != NULL) {
      if (_FP_stricmp (name, iter->name) == 0)
	break;
      iter = iter->NEXT;
    }
    if (_FP_stricmp (name, iter->name) == 0) {
      free (iter->value);
      iter->value = _FP_strdup (value);
    }
    else {
      new->name  = _FP_strdup (name);
      new->value = _FP_strdup (value);
      iter->NEXT = new;
    }
  }
}

void DelInputValues (void)
{
  ifield *iter = inputs, *next;

  while (iter) {
    next = iter->NEXT;

    free (iter->name);
    free (iter->value);
    free (iter);

    iter = next;
  }
  inputs = NULL;
}
    
