/*
 * This file is part of W3Olista, the statistics gathering program for
 * the WWW.                  This sourcecode is (c) Frank Pilhofer '95
 *
 * It is  distributed without warranty. The complete license terms for
 * this software  are included in  the file 'License.html'  in the doc
 * subdirectory. If  your copy came  without this file, you should get
 * it on the Web at    http://www.uni-frankfurt.de/~fp/Tools/W3Olista/
 * or contact the author,               fp@informatik.uni-frankfurt.de
 */

#ifdef SHMONITOR
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include "monitor.h"
#endif

#include <stdio.h>
#include <memory.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "struct.h"
#include "fileio.h"
#include "getcgi.h"
#include "cgiproc.h"
#include "html.h"
#include "project.h"
#include "toolbox.h"
#include "lookup.h"

#ifndef LOGDIR
#define LOGDIR  "/tmp/frank/logs"
#endif
#ifndef LOGFILE
#define LOGFILE "httpd.zeus.access.%b%d%y"
#endif

#ifndef BAROFFSET
#define BAROFFSET 0
#endif
#ifndef BARLENGTH
#define BARLENGTH 0
#endif

#ifndef VERSION
#define VERSION		" NONOFFICIAL "
#endif
#ifndef PATCH
#define PATCH		""
#endif

char   *version   = VERSION;
char   *patch     = PATCH;

char   *logdir    = LOGDIR;
char   *dlogdir   = LOGDIR;
char   *logfile   = LOGFILE;
char   *dlogfile  = LOGFILE;
int     baroffset = BAROFFSET;
int     barlength = BARLENGTH;

char   *servhost  = NULL;
char   *servtype  = NULL;
report *mainreport= NULL;

int     repwhat, maxdepth;
int     linktype, sort;
int     servport, display;
int     resolveaddr;
int     cutprefix;
int     countunique;
struct tm fromdate, todate;

int     keepentries, hostmax;
int     filemax, datemax;

/*
 * arrays of names. must be in same order as the define'd values in
 * project.h
 */
char *repnames[] = {
  "Everything", "Dates",
  "Today", "Yesterday",
  "ThisWeek", "LastWeek",
  "ThisMonth", "LastMonth",
  "Last24Hours", "Last7Days",
  NULL
};

#ifdef SHMONITOR
shmonitor *shmem=NULL;
int        shmid=0;

void shmoninit (void)
{
  struct shmid_ds buf;
  shmem = NULL;

  if ((shmid = shmget (SHMONITOR, sizeof (shmonitor), 0777 | IPC_CREAT)) == -1)
    return;
  if ((shmem = (shmonitor *) shmat (shmid, NULL, 0)) == NULL) {
    if (shmctl (shmid, IPC_STAT, &buf) != -1)
      shmctl (shmid, IPC_RMID, NULL);
    return;
  }
  
  memset (shmem, '\0', sizeof (shmonitor));
  shmem->cookie = 6274;
}

void shmonfin (void)
{
  struct shmid_ds buf;

  if (shmem == NULL)
    return;

  shmdt  ((void *) shmem);
  shmem = NULL;

  if (shmctl (shmid, IPC_STAT, &buf) != -1)
    if (buf.shm_nattch <= 1)
      shmctl (shmid, IPC_RMID, NULL);
}
#endif /* SHMONITOR */


/*
 * I thought I could speed it up by using grep instead of doing every-
 * thing myself. But it didn't.
 */
#ifdef USEEGREP

char *GrepEncode (char *string)
{
  static char result[256];
  unsigned char *ptr=result, *p2=string;

  while (*p2) {
    if ((*p2=='\x20') ||
	(*p2>='\x3a' && *p2<='\x3f') ||
	(*p2>='\x7b')) {
      ptr += sprintf (ptr, "(%c|%02X)", *p2, (int)*p2);
      p2++;
    }
    else
      *ptr++ = *p2++;
  }
  return result;
}

char *InExConstructGrep (char *buffer)
{
  int flag=0, flag2, flag3, count, type;
  char **ptr, temp[256], *bptr=buffer;

  if (!(inhost && *inhost) && !(exhost && *exhost) &&
      !(infile && *infile) && !(exfile && *exfile)) {
    *buffer = '\0';
    return buffer;
  }

  /*
   * Create pipe command to filter all lines that match our
   * Include specifications, if present
   */

  if ((inhost && *inhost) || (infile && *infile)) {
    if (flag++) bptr += sprintf (bptr, " | %s \"", USEEGREP);
    else        bptr += sprintf (bptr, "%s \"",    USEEGREP);

    if ((ptr=inhost) && *inhost) {
      flag2=0;
      bptr += sprintf (bptr, "^[^ ]*");
      while (*ptr) {
	if (flag2++) bptr += sprintf (bptr, "|(%s)", GrepEncode (*ptr));
	else         bptr += sprintf (bptr, "(%s)",  GrepEncode (*ptr));
	ptr++;
      }
      bptr += sprintf (bptr, ".*");
    }
    if ((ptr=infile) && *infile) {
      flag2=0;
      bptr += sprintf (bptr, "GET [^ ]*");
      while (*ptr) {
	if (flag2++) bptr += sprintf (bptr, "|(%s)", GrepEncode (*ptr));
	else         bptr += sprintf (bptr, "(%s)",  GrepEncode (*ptr));
	ptr++;
      }
    }
    bptr += sprintf (bptr, "\"");
  }

  /*
   * Same as above, but throw out all lines that match something
   * we don't want.
   */

  if ((exhost && *exhost) || (exfile && *exfile)) {
    if (flag++) bptr += sprintf (bptr, " | %s -v \"", USEEGREP);
    else        bptr += sprintf (bptr, "%s -v \"",    USEEGREP);

    if ((ptr=exhost) && *exhost) {
      flag2=0;
      bptr += sprintf (bptr, "^[^ ]*");
      while (*ptr) {
	if (flag2++) bptr += sprintf (bptr, "|(%s)", GrepEncode (*ptr));
	else         bptr += sprintf (bptr, "(%s)",  GrepEncode (*ptr));
	ptr++;
      }
      bptr += sprintf (bptr, ".*");
    }
    if ((ptr=exfile) && *exfile) {
      flag2=0;
      bptr += sprintf (bptr, "GET [^ ]*");
      while (*ptr) {
	if (flag2++) bptr += sprintf (bptr, "|(%s)", GrepEncode (*ptr));
	else         bptr += sprintf (bptr, "(%s)",  GrepEncode (*ptr));
	ptr++;
      }
    }
    bptr += sprintf (bptr, "\"");
  }

  /*
   * If we have any file/host reports to do, add more grep
   */

  for (count=0,flag3=0; (type=GetCGIReport (count, temp))!=NOTYPE; count++)
    if (type == HOSTTYPE || type == FILETYPE)
      flag3++;

  if (flag3) {
    if (flag++) bptr += sprintf (bptr, " | %s \"", USEEGREP);
    else        bptr += sprintf (bptr, "%s \"",    USEEGREP);

    for (count=0,flag3=0; (type=GetCGIReport(count,temp))!=NOTYPE; count++) {
      if (type==HOSTTYPE) {
	if (flag3==0) bptr += sprintf (bptr, "^[^ ]*");
	if (flag3++)  bptr += sprintf (bptr, "|(%s)", GrepEncode(*ptr));
	else          bptr += sprintf (bptr, "(%s)",  GrepEncode(*ptr));
      }
    }
    if (flag3) bptr += sprintf (bptr, ".*");

    for (count=0,flag3=0; (type=GetCGIReport(count,temp))!=NOTYPE; count++) {
      if (type==FILETYPE) {
	if (flag3==0) bptr += sprintf (bptr, "GET [^ ]*");
	if (flag3++)  bptr += sprintf (bptr, "|(%s)", GrepEncode(*ptr));
	else          bptr += sprintf (bptr, "(%s)",  GrepEncode(*ptr));
      }
    }
  }
  return buffer;
}

#endif /* USEEGREP */

/*
 * Reads in a line from a file, splits it, and adds it to mainreport.
 * static values are used for values that are needed more than once,
 * so they need only be initialized once. I'm particularly proud of
 * the 'self-modification'-like style of whether a substring or pat-
 * tern is searched. if any string of 'include hosts' etc. is a pat-
 * tern, strmatch() is called. if it is not, strstr is used.
 * call this function with a NULL parameter to reset the static values
 */

static int _int_strstr (char *name, char *str)
{
  return (strstr (name, str) != NULL) ? 1 : 0;
}

void readaction (char *string)
{
  char filename[512], hostname[256], dinfo[256], **ptr, *tptr, *p2;
  static time_t now=0, fromtime=0, totime=0, tt1=0, tt2=0;
  static int (*ihf) (char *, char *) = NULL;
  static int (*ehf) (char *, char *) = NULL;
  static int (*iff) (char *, char *) = NULL;
  static int (*eff) (char *, char *) = NULL;
  static struct tm dnow, dt1;
  static int count=0;
  int status, size, diff;
  struct tm date;
  time_t then;

  if (string==NULL) { /* reset values */
    now = fromtime = totime = tt1 = tt2 = 0;
    ihf = ehf = iff = eff = NULL;
    count = 0;
  }
#ifdef SHMONITOR
  if ((++count % 500) == 0) {
    if (shmem && !shmem->haschanged) {
      shmem->flag    = MF_LOADING;
      shmem->totline = count;
      shmem->momline = momlogline;
      shmem->heapused= heapused;

      strcpy (shmem->momfile, momlogfile);
      shmem->haschanged = 1;
    }
  }
#endif

  if (scanstr (string, hostname, filename, dinfo, &status, &size) == 0)
    return;

  if (dtreetotm (&date, dinfo) == NULL)
    return;

  if (status != 200)
    return;

  _FP_stolower (hostname);

  if (now==0) {
    now = time(NULL);
#ifdef GMT
    memcpy (&dnow, gmtime (&now), sizeof (struct tm));
#else
    memcpy (&dnow, localtime (&now), sizeof (struct tm));
#endif
  }

  /*
   * test if we have to report this entry by checking its timestamp
   * if we don't like the timestamp, we return immediately
   * if we break out of this switch, then timestamp is OK
   */

  switch (repwhat) {
  case REVERYTHING:  /* report _everything_ */
    break;
  case RLAST24HOURS: /* report everything from the last 24 hours */
    /*
     * aargh! why doesn't this work? These damn timezones!
     */
    /*
     * if ((then = mktime (&date)) == (time_t)-1)
     *   return;
     * else if (now-then > 86400)
     *  return;
     * break;
     */
    /*
     * let's try a different approach. Ugly, isn't it?
     */
    if (  (date.tm_year  == dnow.tm_year && date.tm_yday   == dnow.tm_yday) ||
	(((date.tm_year  == dnow.tm_year && date.tm_yday+1 == dnow.tm_yday) ||
	  (date.tm_year+1== dnow.tm_year && dnow.tm_yday   == 0)           ) &&
	 ((  date.tm_hour >  dnow.tm_hour) ||
	  (  date.tm_hour == dnow.tm_hour &&
	   ((date.tm_min  >  dnow.tm_min) ||
	    (date.tm_min  == dnow.tm_min && date.tm_sec >= dnow.tm_sec))))))
      ;
    else
      return;
    break;

  case RTHISWEEK:    /* from last sunday up to today */
    if (tt1==0) {
      memcpy (&dt1, &dnow, sizeof (struct tm));
      dt1.tm_mday -= dt1.tm_wday;
      dt1.tm_hour  = 0;
      dt1.tm_min   = 0;
      dt1.tm_sec   = 0;
      dt1.tm_isdst =-1;
      tt1 = mktime (&dt1);
    }
    if ((then = mktime (&date)) == (time_t)-1)
      return;
    if (then < tt1)
      return;
    break;
  case RLASTWEEK:    /* from sun last week to sat last week */
    if (tt1==0) {
      memcpy (&dt1, &dnow, sizeof (struct tm));
      dt1.tm_mday -= dt1.tm_wday;
      dt1.tm_hour  = 0;
      dt1.tm_min   = 0;
      dt1.tm_sec   = 0;
      dt1.tm_isdst =-1;
      tt1 = mktime (&dt1);
      tt2 = tt1 - 7 * 86400;
    }
    if ((then = mktime (&date)) == (time_t)-1)
      return;
    if (then < tt2 || then >= tt1)
      return;
    break;
  case RTHISMONTH:  /* from 1st of this month to today */
    if ((date.tm_mon  != dnow.tm_mon)   ||
	(date.tm_year != dnow.tm_year))
      return;
    break;
  case RLASTMONTH:  /* the whole last month */
    if ((date.tm_mon+1 == dnow.tm_mon && date.tm_year == dnow.tm_year) ||
	(date.tm_mon   == 11 &&
	 dnow.tm_mon   ==  0 &&
	 date.tm_year+1== dnow.tm_year))
      ;
    else
      return;
    break;
  case RTODAY:      /* today */
    if (date.tm_mday == dnow.tm_mday &&
	date.tm_mon  == dnow.tm_mon  &&
	date.tm_year == dnow.tm_year)
      ;
    else
      return;
    break;
  default:          /* common processing for other cases */
    date.tm_hour = dnow.tm_hour;
    date.tm_min  = dnow.tm_min;
    date.tm_sec  = dnow.tm_sec;
    
    if ((then = mktime (&date)) == (time_t)-1)
      return;

    /*
     * we add 7200 sec here because of possible DST switching, when
     * one day is shorter than 24 hours
     */

    diff = (now - then + 7200) / 86400;
    
    switch (repwhat) {
    case RYESTERDAY:/* yesterday */
      if (diff != 1)
	return;
      break;
    case RLAST7DAYS:/* last 7 days. this excludes today. */
      if (diff == 0 || diff > 7)
	return;
      break;
    case RDATES:    /* a selected range of dates */
      if (fromtime==0) {
	fromtime = mktime (&fromdate);
	totime   = mktime (&todate);
      }
      if ((then < fromtime) || (then > totime))
	return;
      break;
    } /* inner switch */
    break;
  } /* main switch */

  /*
   * maybe grep has already done all the filtering
   */

#ifndef USEEGREP

  if (ihf==NULL) {
    if (inhost && *inhost) {
      if (strpbrk (*inhost, "*?") != NULL)
	ihf = _FP_strmatch;
      else
	ihf = _int_strstr;
    }
    else /* if we don't need it anyway, set to anything non-NULL as marker */
      ihf = (int (*) (char *, char *)) -1;

    if (exhost && *exhost) {
      if (strpbrk (*exhost, "*?") != NULL)
	ehf = _FP_strmatch;
      else
	ehf = _int_strstr;
    }
    if (infile && *infile) {
      if (strpbrk (*infile, "*?") != NULL)
	iff = _FP_strmatch;
      else
	iff = _int_strstr;
    }
    if (exfile && *exfile) {
      if (strpbrk (*exfile, "*?") != NULL)
	eff = _FP_strmatch;
      else
	eff = _int_strstr;
    }
  }

  if ((ptr=inhost) && *inhost) {
    while (*ptr) {
      if ((*ihf) (hostname, *ptr))
	break;
      ptr++;
    }
    if (*ptr == NULL)
      return;
  }

  if ((ptr=exhost) != NULL) {
    while (*ptr) {
      if ((*ehf) (hostname, *ptr))
	return;
      ptr++;
    }
  }
  
  if ((ptr=infile) && *infile) {
    while (*ptr) {
      if ((*iff) (filename, *ptr))
	break;
      ptr++;
    }
    if (*ptr == NULL)
      return;
  }

  if ((ptr=exfile) != NULL) {
    while (*ptr) {
      if ((*eff) (filename, *ptr))
	return;
      ptr++;
    }
  }
#endif

  /*
   * try to resolve a numerical address. Terribly slow, since 
   * it obviously doesn't cache failures. Do I have to hash them
   * on my own? Yes, I did have to, so it's a little faster now.
   */

#ifdef RESOLVEADDR
  if (resolveaddr && atoi (hostname)>0) {
    if ((tptr = LookupDomain (hostname)) != NULL)
      strcpy (hostname, tptr);
  }
#endif

  /*
   * combine all numerical addresses under the (num) top-level domain.
   * they'll be specially handled on output.
   */
  if (atoi (hostname) > 0)
    strcat (hostname, ".(num)");

  /*
   * if this is a logentry for myself (the olista program), then cut off
   * all those weird extra-path information that would show up otherwise
   * Yes, I know, it's not state-of-the-art.
   */
  if ((tptr = strstr (filename, "/olista/")) != NULL)
    *(tptr + 6) = '\0';

  /*
   * sometimes, we're getting malformed addresses (seem to be caused by
   * the local netscape installation) still with complete http:// prefix.
   * cut it off. But if we should want to check proxy or cache files, we
   * might need it.
   */

  if (cutprefix) {
    p2 = filename;
    if (strncmp (filename, "http://", 7) == 0)
      tptr = filename + 7;
    else if (strncmp (filename, "ftp://", 6) == 0)
      tptr = filename + 6;
    else if (strncmp (filename, "gopher://", 9) == 0)
      tptr = filename + 9;
    else if (strncmp (filename, "wais://", 7) == 0)
      tptr = filename + 7;
    else
      tptr = NULL;
    if (tptr) {
      while (*tptr && *tptr!='/')
	tptr++;
      if (*tptr != '/')
	return;
      while (*tptr)
        *p2++ = *tptr++;
      *p2 = '\0';
    }
  }
  
  /*
   * if we got through to here, then this is an entry we want to report
   */

  addentry (mainreport, hostname, filename, dinfo, size, 
	    hostmax, filemax, datemax, keepentries);
}

/*
 * checks whether we want to use a logfile or not. called from fileio.c
 * we have to reverse the action of strftime() here. getdate() does the
 * same, but needs tampering with environments.
 * this function is only called once per logfile, so it's hardly worth
 * optimizing here. But some intelligence really pays off; each logfile
 * we don't need to read saves precious time.
 *
 * strtok() would be ideal to parse multiple logfile patterns, but it's
 * already called by readlofiles() and numlogfiles(). aargh!
 */

int acceptfile (char *filename)
{
  int index, result, error=0, diff, info=0, yinfo=0, minfo=0, dinfo=0, igf;
  char *p1=NULL, *p2, temp[256], *cfptr, c;
  struct tm date, rfdate, rtdate, dnow, *dtmp;
  time_t then, now, t1, t2;

  memset (&date,  '\0', sizeof (struct tm));
  now    = time(NULL);
  result = LOGNOLOG;
  cfptr  = logfile;

  memcpy (&dnow, localtime (&now), sizeof (struct tm));

  date.tm_year = dnow.tm_year;
  date.tm_mon  = dnow.tm_mon;
  date.tm_mday = dnow.tm_mday;

 afretry:
  while (cfptr && *cfptr) {
    p1 = filename;
    p2 = cfptr;

    while (*p2 && *p2 != ';' && !error) {
      if (*p2 == '.') {
	if (*p1 == '\0') {
	  error = 1;
	  break;
	}
	p1++; p2++;
	continue;
      }
      else if (*p2 != '%') {
	if (*p1 != *p2) {
	  error = 1;
	  break;
	}
	p1++; p2++;
	continue;
      }
      if (*++p2 == '*') {
	p2++;
	igf=1;
      }
      else
	igf=0;
      
      /*
       * process % sequence
       */
      
      switch (*p2) {
      case '%':
	if (*p1 != *p2)
	  error = 1;
	p1++;
	break;
      case 'b':       /* 3 character short english month, first one initial */
	for (index=1; mnames[index]; index++)
	  if (_FP_strnicmp (p1, mnames[index], strlen (mnames[index])) == 0)
	    break;
	if (mnames[index]) {
	  date.tm_mon = index - 1;
	  p1 += strlen (mnames[index]);
	  if (!igf) {
	    info  = 1;
	    minfo = 1;
	  }
	}
	else
	  error = 1;
	break;
      case 'B':       /* long english month name. first character initial */
	for (index=1; mlnames[index]; index++)
	  if (_FP_strnicmp (p1, mlnames[index], strlen (mlnames[index])) == 0)
	    break;
	if (mlnames[index]) {
	  date.tm_mon = index - 1;
	  p1 += strlen (mlnames[index]);
	  if (!igf) {
	    info  = 1;
	    minfo = 1;
	  }
	}
	else
	  error = 1;
	break;
      case 'd':       /* day of the month. exactly two digits */
	if (*p1 && isdigit (*p1) && isdigit (*(p1+1))) {
	  date.tm_mday = 10 * (*p1 - '0') + (*(p1+1) - '0');
	  p1          += 2;
	}
	else if (*p1 && isdigit (*p1)) {
	  date.tm_mday = *p1 - '0';
	  p1          += 1;
	}
	else
	  error = 1;

	if (!error && !igf) {
	  info  = 1;
	  dinfo = 1;
	}
	break;
      case 'm':       /* numerical month. exactly two digits */
	if (*p1 && isdigit (*p1) && isdigit (*(p1+1))) {
	  date.tm_mon  = 10 * (*p1 - '0') + (*(p1+1) - '0') - 1;
	  p1          += 2;
	}
	else if (*p1 && isdigit (*p1)) {
	  date.tm_mon  = *p1 - '0' - 1;
	  p1          += 1;
	}
	else
	  error = 1;

	if (!error && !igf) {
	  info  = 1;
	  minfo = 1;
	}
	break;
      case 'y':      /* short numerical year. exactly two digits */
	if (*p1 && isdigit (*p1) && isdigit (*(p1+1))) {
	  date.tm_year = 10 * (*p1 - '0') + (*(p1+1) - '0');
	  p1          += 2;
	}
	else
	  error = 1;

	if (date.tm_year < 60)
	  date.tm_year += 100;

	if (!error && !igf) {
	  info  = 1;
	  yinfo = 1;
	}
	break;
      case 'Y':      /* long numerical year. exactly four digits */
	if (strlen(p1)>=4) {
	  c = *(p1 + 4); *(p1 + 4) = '\0';
	  date.tm_year = atoi (p1) - 1900;
	  p1 += 4; *p1 = c;
	}
	else
	  error = 1;

	if (!error && !igf) {
	  info  = 1;
	  yinfo = 1;
	}
	break;
      default:
	error = 1;
      }
      p2++;
    }
    if (*p2!='\0' && *p2!=';')
      error = 1;

    if (!error)
      break;

    while (*cfptr && *cfptr!=';' && *cfptr!=' ')
      cfptr++;
    while (*cfptr==';' || *cfptr==' ')
      cfptr++;
    error = 0;
    continue;
  }

  if (p1==NULL || error)
    return 0;

  if (*p1 == '\0')
    result = LOGPLAIN;
  else if (_FP_strnicmp (p1, ".z",  2) == 0)
    result = LOGCOMPRESSED;
  else if (_FP_strnicmp (p1, ".gz", 3) == 0)
    result = LOGGZIPPED;
  else if (*p2==';' || *p2==' ') {
    while (*cfptr && *cfptr!=';' && *cfptr!=' ')
      cfptr++;
    while (*cfptr==';' || *cfptr==' ')
      cfptr++;
    goto afretry;
  }
  else
    return 0;

  if (info == 0) /* No Date info in filename, so use it */
    return result;

  /*
   * if there's any date report pending, it might help us immensely
   * to filter files if we should report on a complete month or more
   */
  if (repwhat==REVERYTHING || repwhat==RTHISMONTH ||
      repwhat==RLASTMONTH  || repwhat==RDATES) {
    for (index=0,igf=0; (igf=GetCGIReport(index, temp))!=NOTYPE; index++) {
      if (igf==DATETYPE)
	break;
    }
    if (igf==DATETYPE && dinfo && minfo && yinfo) {
      if (dtreetotm (&rfdate, cgidecode (temp))) {
	memcpy (&rtdate, &rfdate, sizeof (struct tm));
	date.tm_isdst = -1;
	if ((then = mktime (&date)) == (time_t)-1)
	  return 0;
	if (mktime(&rtdate)-then < 86400)
	  result=LOGNOLOG;
	return result;
      }
    }
  } 
  
  /*
   * if we have to process everything anyway, skip all the testing.
   */

  if (repwhat==REVERYTHING)
    return result;

  /*
   * Check what information we have, and use this info to decide whether
   * this logfile contains useful information
   */

#ifdef DEBUG
  fprintf (stderr, "acceptfile() dinfo=%d minfo=%d yinfo=%d on %s\n",
	   dinfo, minfo, yinfo, filename);
#endif

  if (dinfo && minfo && yinfo) { /* we know everything */
    info          =  1;
    date.tm_isdst = -1;

    date.tm_hour = dnow.tm_hour;
    date.tm_min  = dnow.tm_min;
    date.tm_sec  = dnow.tm_sec;
    
    if ((then = mktime (&date)) == (time_t)-1)
      return 0;

    /*
     * we add 7200 secs here because of possible DST switching, when
     * one day is shorter than 24 hours
     */

    diff = (now - then + 7200) / 86400;

    switch (repwhat) {
    case RTODAY:
      if (diff != 0) 
	result = LOGNOLOG;
      break;
    case RYESTERDAY:
      if (diff != 1)
	result = LOGNOLOG;
      break;
    case RTHISWEEK:
      if (diff > dnow.tm_wday)
      result = LOGNOLOG;
      break;
    case RLASTWEEK:
      if (diff <=   dnow.tm_wday ||
	  diff  > 7+dnow.tm_wday)
	result = LOGNOLOG;
      break;
    case RTHISMONTH:
      if ((date.tm_mon  != dnow.tm_mon)  ||
	  (date.tm_year != dnow.tm_year))
	result = LOGNOLOG;
      break;
    case RLASTMONTH:
      if ((date.tm_mon+1  == dnow.tm_mon && date.tm_year == dnow.tm_year) ||
	  (date.tm_mon    == 11 &&
	   dnow.tm_mon    ==  0 &&
	   date.tm_year+1 == dnow.tm_year))
	;
      else
	result = LOGNOLOG;
      break;
    case RLAST24HOURS: /* this one needs today's and yesterday's log */
      if (diff > 1)
	result = LOGNOLOG;
      break;
    case RLAST7DAYS:
      if (diff == 0 || diff > 7)
	result = LOGNOLOG;
      break;
    case RDATES:
      if ((then < mktime (&fromdate)) || (then > mktime (&todate)))
	result=LOGNOLOG;
      break;
    default:
      result = LOGNOLOG;
      break;
    } /* switch */
  }
  else if (minfo && yinfo) { /* we can't get day info from the filename  */
    switch (repwhat) {
    case RTODAY:
      if (date.tm_mon  == dnow.tm_mon && date.tm_year == dnow.tm_year);
      else result = LOGNOLOG;
      break;
    case RYESTERDAY:
      then = now - 86400;
      dtmp = localtime (&then);
      if (date.tm_mon  == dtmp->tm_mon && date.tm_year == dtmp->tm_year);
      else result = LOGNOLOG;
      break;
    case RTHISWEEK:
      then = now - 86400 * dnow.tm_wday;
      dtmp = localtime (&then);
      if ((date.tm_mon == dnow.tm_mon  && date.tm_year == dnow.tm_year) ||
	  (date.tm_mon == dtmp->tm_mon && date.tm_year == dtmp->tm_year));
      else result = LOGNOLOG;
      break;
    case RLASTWEEK:
      then = now  - 86400 * (dnow.tm_wday + 7);
      memcpy (&rfdate, localtime (&then), sizeof (struct tm));
      then = then - 86400 * 7 + 1;
      memcpy (&rtdate, localtime (&then), sizeof (struct tm));
      
      if ((date.tm_mon==rfdate.tm_mon && date.tm_year==rfdate.tm_year) ||
	  (date.tm_mon==rtdate.tm_mon && date.tm_year==rfdate.tm_year));
      else result = LOGNOLOG;
      break;
    case RTHISMONTH:
      if (date.tm_mon == dnow.tm_mon && date.tm_year == dnow.tm_year);
      else result = LOGNOLOG;
      break;
    case RLASTMONTH:
      if ((date.tm_mon+1  == dnow.tm_mon && date.tm_year == dnow.tm_year) ||
	  (date.tm_mon    == 11 &&
	   dnow.tm_mon    ==  0 &&
	   date.tm_year+1 == dnow.tm_year))
	;
      else result = LOGNOLOG;
      break;
    case RLAST24HOURS:
      then = now - 86400;
      dtmp = localtime (&then);
      if ((date.tm_mon  == dtmp->tm_mon && date.tm_year == dtmp->tm_year) ||
	  (date.tm_mon  == dnow.tm_mon  && date.tm_year == dnow.tm_year));
      else result = LOGNOLOG;
      break;
    case RLAST7DAYS:
      then = now - 86400;
      memcpy (&rfdate, localtime (&then), sizeof (struct tm));
      then = then - 86400 * 7 + 1;
      memcpy (&rtdate, localtime (&then), sizeof (struct tm));
      if ((date.tm_mon==rfdate.tm_mon && date.tm_year==rfdate.tm_year) ||
	  (date.tm_mon==rtdate.tm_mon && date.tm_year==rfdate.tm_year));
      else result = LOGNOLOG;
      break;
    case RDATES:
      date.tm_isdst = -1;
      date.tm_mday  =  1;
      date.tm_hour  = date.tm_min = date.tm_sec = 0;
      if ((t1 = mktime (&date)) == (time_t) -1)
	return LOGNOLOG;
      date.tm_mon += 1;
      if ((t2 = mktime (&date)) == (time_t) -1)
	return LOGNOLOG;
      t2 = t2 - 1;

#ifdef DEBUG
      fprintf (stderr, "acceptfile() RDATES t1=%d t2=%d fromdate=%d todate=%d\n", t1, t2, mktime(&fromdate), mktime(&todate));
#endif

      if ((t1 < mktime (&fromdate) && t2 < mktime (&todate)) ||
	  (t1 > mktime (&fromdate) && t2 > mktime (&todate)))
	result = LOGNOLOG;
      break;
    default:
      result = LOGNOLOG;
      break;
    } /* switch */
  }
  else { /* we have only the year info. I don't bother checking again */
    ;
  }

  return result;
}

report *doreports (FILE *logfile, report *initial)
{
  report *result=NULL, *new=NULL, *temp;
  char gstring[256], string[256];
  int type, index=0;
  void *test;

  while (1) {
    if ((type = GetCGIReport (++index, gstring)) == NOTYPE)
      break;
    strcpy (string, cgidecode (gstring));

    if (new==NULL)
      temp = initial;
    else
      temp = new;

    switch (type) {
    case FILETYPE:
      test = locatefile (temp, string);
      break;
    case HOSTTYPE:
      test = locatehost (temp, string);
      break;
    case DATETYPE:
      test = locatedate (temp, string);
      break;
    default:
      test = NULL;
    }

    if (test==NULL)
      break;

#ifdef SHMONITOR
    if (shmem && !shmem->haschanged) {
      shmem->flag    = MF_REPORT;
      shmem->reptype = type;
      strcpy (shmem->momfile, string);
      shmem->haschanged = 1;
    }
#endif

    if ((result = newreport (test, type)) == NULL) {
      if (new)
	killreport (new);
      return NULL;
    }

    if (new)
      killreport (new);

    new = result;
  }

  if (result == NULL)
    return initial;

  return result;
}

int main (int argc, char *argv[])
{
  char *p1, *p2, temp[256], mlogdir[256], *usegrep=NULL;
  int iscgi, files, lines, iscopy=0;
  report *result=NULL;
#ifdef USEEGREP
  char grepstring[1024];
#endif
#ifdef RESOLVEADDR
  char cachefile[256];
  int cfro, cfip, cflu;
#endif

  /*
   * Just done so I'll get the CoreDumps :-))
   */

  chdir  ("/usr/usersb2/w93/fp/c/olista");

  /*
   * what shall we do if initialization fails?
   */

  if (InitInputValues (NULL, argc, argv) != _CGI_OK) {
    if (getenv ("SCRIPT_NAME") == NULL) { /* Plain Text */
      printf ("%s: Fatal error during startup. This really shouldn't happen.\n", argv[0]);
    }
    else {
      fprintf (stdout, "Content-type: text/html\n\n\
<html><head>\n\
<title>W3Olista Error</title>\n\
</head><body>\n\
<h1>W3Olista couldn't startup</h1>\n\
If you see this, then something is seriously wrong. This really shouldn't\n\
happen.\n\
</body></html>\n");
    }
    return 0;
  }

  /*
   * Assume we're CGI here, or at least HTML.
   */

  if (1 /* GetInputValue ("cgi") || GetInputValue ("html") */ )
    iscgi = 1;
  else
    iscgi = 0;

#ifdef SHMONITOR
  shmoninit ();

  if (shmem) {
    shmem->flag = MF_INIT;

    if ((p1 = GetInputValue ("Remote")) != NULL)
      strcpy (shmem->remote, p1);
    else
      *(shmem->remote) = '\0';

    shmem->haschanged = 1;
  }
#endif

  /*
   * I don't need a fancy sighandler when debugging
   */

#ifndef DEBUG
  if (iscgi) {
    signal (SIGILL, HTMLSigHandler);
    signal (SIGBUS, HTMLSigHandler);
    signal (SIGSEGV,HTMLSigHandler);
    signal (SIGTERM,HTMLSigHandler);
  }
#endif

  if ((p1 = GetInputValue ("logdir")) != NULL) {
    strcpy (mlogdir, cgidecode (p1));
    logdir = mlogdir;
  }
  if ((p1 = GetInputValue ("logfile")) != NULL)
    logfile = p1;

  maxdepth = 0;
  if ((p1 = GetInputValue ("HostDetail")) != NULL)
    maxdepth |= atoi (p1) << HOSTDETOFF;
  if ((p1 = GetInputValue ("FileDetail")) != NULL)
    maxdepth |= atoi (p1) << FILEDETOFF;
  if ((p1 = GetInputValue ("DateDetail")) != NULL)
    maxdepth |= atoi (p1) << DATEDETOFF;

#ifdef DEBUG
  fprintf (stderr, "logdir=%s logfile=%s Detail=%d\n", logdir, logfile, maxdepth);
#endif

  display = 0;
  if ((p1 = GetInputValue ("HostList")) != NULL) {
    if (*p1=='\0' || _FP_stricmp (p1, "yes") == 0) display |= HOSTLIST;
    else if (_FP_strnicmp (p1, "sum", 3) == 0)     display |= HOSTSUM;
  }
  if ((p1 = GetInputValue ("FileList")) != NULL) {
    if (*p1=='\0' || _FP_stricmp (p1, "yes") == 0) display |= FILELIST;
    else if (_FP_strnicmp (p1, "sum", 3) == 0)     display |= FILESUM;
  }
  if ((p1 = GetInputValue ("DateList")) != NULL) {
    if (*p1=='\0' || _FP_stricmp (p1, "yes") == 0) display |= DATELIST;
    else if (_FP_strnicmp (p1, "sum", 3) == 0)     display |= DATESUM;
  }
  if ((p1 = GetInputValue("EntryList")) != NULL) {
    if (*p1=='\0' || _FP_stricmp (p1, "yes") == 0) display  = ENTRYLIST;
  }

  if ((p1 = GetInputValue ("Sort")) != NULL) {
    if (_FP_stricmp (p1, "by-alpha") == 0)     sort = SORTALPHA;
    else if (_FP_stricmp (p1, "by-req")  == 0) sort = SORTREQ;
    else if (_FP_stricmp (p1, "by-size") == 0) sort = SORTSIZE;
    else sort = SORTALPHA;
  }
  else
    sort = SORTALPHA;

  if ((p1 = GetInputValue ("Report")) != NULL) {
    for (repwhat=0; repnames[repwhat]; repwhat++)
      if (_FP_stricmp (p1, repnames[repwhat]) == 0)
	break;
    if (repnames[repwhat] == NULL)
      repwhat = RTODAY;
  }
  else
    repwhat = RTODAY;

#ifdef DEBUG
  fprintf (stderr, "Display=%d Sort=%d Report=%d\n",
	   display, sort, repwhat);
#endif

  if (repwhat == RDATES) {
    if ((p1 = GetInputValue ("FromDate")) != NULL) {
      if (ctodate (&fromdate, cgidecode (p1))) {
	if ((p1 = GetInputValue ("ToDate")) != NULL) {
	  if (ctodate (&todate, cgidecode (p1)) == NULL)
	    memcpy (&todate, &fromdate, sizeof (struct tm));
	}
	else 
	  memcpy (&todate, &fromdate, sizeof (struct tm));

	if (todate.tm_hour == 0 && todate.tm_min == 0) {
	  todate.tm_hour = 23; /* we want to catch everything of the last */
	  todate.tm_min  = 59; /* date                                    */
	  todate.tm_sec  = 59;
	}
      }
      else repwhat = RTODAY;
    }
    else repwhat = RTODAY;
  }

#ifdef DEBUG
  if (repwhat == RDATES)
    fprintf (stderr, "Dates FromDate=%02d.%02d.%02d ToDate=%02d.%02d.%02d\n",
	     fromdate.tm_mday, fromdate.tm_mon+1, fromdate.tm_year,
	     todate.tm_mday,   todate.tm_mon+1,   todate.tm_year);
#endif

  if ((p1 = GetInputValue ("Link")) != NULL) {
    if (_FP_stricmp (p1, "NoLink") == 0)          linktype = LNOLINK;
    else if (_FP_stricmp (p1, "RepLink") == 0)    linktype = LREPLINK;
    else if (_FP_stricmp (p1, "FileLink") == 0) {
      if ((servhost = GetInputValue ("Server")) == NULL) {
	if ((servhost = GetInputValue ("SERVER_NAME")) == NULL)
	  linktype = LNOLINK;
	else
	  linktype = LFILELINK;
      }
      else
	linktype = LFILELINK;
      if (linktype == LFILELINK) {
	if ((servtype = GetInputValue ("ServerType")) == NULL)
	  servtype = "http";
	if ((p2 = GetInputValue ("ServerPort")) == NULL)
	  servport = 80;
	else if ((servport = atoi (p2)) == 0)
	  servport = 80;
      }
    }
    else
      linktype = LNOLINK;
  }
  else
    linktype = LREPLINK;

  if (linktype==LREPLINK && GetInputValue("SCRIPT_NAME")==NULL)
    linktype = LNOLINK;

  if ((p1 = GetInputValue ("BarLength")) != NULL)
    if ((barlength = atoi (p1)) <= 0)
      barlength = BARLENGTH;
  if ((p1 = GetInputValue ("BarOffset")) != NULL)
    if ((baroffset = atoi (p1)) <= 0)
      baroffset = BAROFFSET;
  if ((p1 = GetInputValue ("CutPrefix")) != NULL) {
    if (*p1=='\0' || _FP_stricmp(p1, "yes")==0)
      cutprefix = 1;
    else
      cutprefix = 0;
  }
  else
    cutprefix = 1;
  if ((p1 = GetInputValue ("CountUnique")) != NULL) {
    if (*p1=='\0' || _FP_stricmp(p1, "yes")==0)
      countunique = 1;
    else
      countunique = 0;
  }
  else
    countunique = 0;

#ifdef RESOLVEADDR
  if ((p1 = GetInputValue ("ResolveAddr")) != NULL) {
    if (*p1=='\0' || _FP_stricmp(p1, "yes")==0)
      resolveaddr = 1;
    else
      resolveaddr = 0;

    if ((p1 = GetInputValue ("ResolveCacheFile")) != NULL) {
      cfro = 0;
      strcpy (cachefile, p1);
    }
    else if ((p1 = GetInputValue ("ResolveCacheFileReadOnly")) != NULL) {
      cfro = 1;
      strcpy (cachefile, p1);
    }
    else
      cachefile[0] = '\0';

    if ((p1 = GetInputValue ("ResolveCacheLogIP")) != NULL) {
      if (*p1=='\0' || _FP_stricmp (p1, "yes")==0)
	cfip = 1;
      else
	cfip = 0;
    }
    else
      cfip = 1;

    if ((p1 = GetInputValue ("ResolveCacheLookUp")) != NULL) {
      if (*p1=='\0' || _FP_stricmp (p1, "yes")==0)
	cflu = 1;
      else
	cflu = 0;
    }
    else
      cflu = 1;
  }
  else
    resolveaddr = 0;
#else
  resolveaddr = 0;
#endif

  GetIncludeExclude ();

  if ((p1 = GetInputValue ("form")) || GetInputValue ("sendcomment")) {
    htmlheader (stdout);

    if (p1)
      PrintForm    (stdout);
    else
      PrintSendCom (stdout);

    htmlfooter (stdout);
    DelIncludeExclude ();
#ifdef SHMONITOR
    shmonfin ();
#endif
    return 0;
  }
#ifdef USEEGREP
  usegrep = InExConstructGrep (grepstring);
#endif

  if (display == 0) {
    if (iscgi) {
      htmlheader (stdout);
      printf     ("<h1>Error</h1>\nNothing to display.<br>\n");
      htmlfooter (stdout);
    }
    else
      printf ("Error! Nothing to display.\n");
    
    goto abort;
  }

#ifdef RESOLVEADDR
  if (resolveaddr)
    InitLookup (cachefile[0]?cachefile:NULL, cfro, cfip, cflu);
#else
  if (resolveaddr)
    InitLookup (NULL, 1, 0, 0); /* this is a dummy anyway */
#endif

  if ((mainreport = newreport (NULL,0)) == NULL) {
    if (iscgi) {
      htmlheader (stdout);
      printf     ("<h1>Error</h1>\nCould not create empty report.<br>\n");
      htmlfooter (stdout);
    }
    else
      printf ("Error! Could not create empty report.\n");

    goto abort;
  }

#ifdef SHMONITOR
  if ((files = numlogfiles (logdir, usegrep, acceptfile, &lines)) == 0) {
#else
  if ((files = numlogfiles (logdir, usegrep, acceptfile, NULL)) == 0) {
#endif
    if (iscgi) {
      htmlheader (stdout);
      printf     ("<h1>Error</h1>\nNo logs found for this date.<br>\n");
      htmlfooter (stdout);
    }
    else
      printf ("Error! No logs found for this date.\n");

    goto abort;
  }

#ifdef SHMONITOR
  if (shmem)
    shmem->hopline = lines;
#endif

  /*
   * decide if we need entries, and to which depth we keep info
   * trees. Looks like bad design, having to substract different
   * values from the display depth
   */
  if (!iscgi || GetCGIReport (1, temp) != NOTYPE || display==ENTRYLIST) {
    keepentries =  1;
    hostmax     = -1;
    filemax     = -1;
    datemax     = -1;
  }
  else {
    keepentries = 0;

    if (countunique)
      hostmax = -1;
    else if ((display&HOSTLIST) == HOSTLIST)
      hostmax = (HOSTDEPTH(maxdepth)==0) ? -1 : HOSTDEPTH(maxdepth);
    else if ((display&HOSTLIST) == HOSTSUM)
      hostmax = 1;
    else
      hostmax = 0;
    
    if ((display&FILELIST) == FILELIST)
      filemax = (FILEDEPTH(maxdepth)==0) ? -1 : FILEDEPTH(maxdepth);
    else if ((display&FILELIST) == FILESUM)
      filemax = 2;
    else
      filemax = 0;
    
    if ((display&DATELIST) == DATELIST)
      datemax = (DATEDEPTH(maxdepth)==0) ? -1 : DATEDEPTH(maxdepth);
    else if ((display&DATELIST) == DATESUM)
      datemax = 2;
    else
      datemax = 0;
  }

  readlogfiles (logdir, usegrep, acceptfile, readaction);

  if ((result = doreports (stdout, mainreport)) == NULL) {
    if (iscgi) {
      htmlheader (stdout);
      printf     ("<h1>Error</h1>\ndoreports() returned NULL value.<br>\n");
      htmlfooter (stdout);
    }
    else
      printf     ("Error! doreports() returned NULL value.\n");

    goto abort;
  }
  if (result == mainreport)
    iscopy = 1;

  result = sortbyrs (result, 0, sort);

  if (iscgi) {
#ifdef SHMONITOR
    if (shmem && !shmem->haschanged) {
      shmem->flag       = MF_WRITE;
      shmem->haschanged = 1;
    }
#endif
    CGImain (result, stdout);
    goto abort;
  }

 abort:
  if (resolveaddr) /* Kill hashtable */
    DestroyLookup ();

  if (!iscopy)
    killreport (result);

  killreport (mainreport);
  DelIncludeExclude ();
  DelInputValues ();

#ifdef SHMONITOR
  shmonfin ();
#endif

  return 0;
}
