/*
 * This file is part of W3Olista, the statistics gathering program for
 * the WWW.                  This sourcecode is (c) Frank Pilhofer '95
 *
 * It is  distributed without warranty. The complete license terms for
 * this software  are included in  the file 'License.html'  in the doc
 * subdirectory. If  your copy came  without this file, you should get
 * it on the Web at    http://www.uni-frankfurt.de/~fp/Tools/W3Olista/
 * or contact the author,               fp@informatik.uni-frankfurt.de
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "struct.h"
#include "toolbox.h"

/*
 * add info to global structs. Welcome to Pointermania!
 */

int heapused = 0;

/*
 * Local functions that are not exported
 */

char *		tmtodtree	(struct tm *);
struct tm *	dtreetotm	(struct tm *, char *);

static host *	addhost_r	(report *, char *, host *, host *, long, int, int);
static file *	addfile_r	(report *, char *, file *, file *, long, int, int);
static dtree *	adddate_r	(report *, char *, dtree *, dtree *,long,int, int);
static entries *addentries	(entries *, entry *);
static report *	newreport_r	(report *, void *, int);
static entry *	sortentry	(entry *, int);
static void *	sortbyrs_r	(void *, int, int);

static void 	killentries	(entries *);
static void 	killhost	(host *);
static void 	killfile	(file *);
static void	killdtree	(dtree *);

/*
 * Month mapping
 */
char *mnames[] = {
  "",
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
  NULL
};

char *mlnames[] = {
  "",
  "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November", "December",
  NULL
};

/*
 * Empty requests upon a directory
 */

#ifdef WELCOME
static char *welcome = WELCOME;
#else
static char *welcome = "welcome.html";
#endif

/*
 * Macros
 */

#define addhost(r,s,z,m)	addhost_r(r,s,r->hlist,NULL,z,0,m);
#define addfile(r,s,z,m)	addfile_r(r,s,r->flist,NULL,z,0,m);
#define adddate(r,s,z,m)	adddate_r(r,s,r->dlist,NULL,z,0,m);

/*
 * Implementation 
 */

/*
 * Helper functions for date conversion
 */

char *tmtodtree (struct tm *date)
{
  static char result[256];

  sprintf (result, "%d %s %s %d %d %d %d",
	   date->tm_year + 1900,
	   mnames[date->tm_mon + 1],
	   mnames[date->tm_mon + 1],
	   date->tm_mday,
	   date->tm_hour,
	   date->tm_min,
	   date->tm_sec);

  return result;
}

struct tm *dtreetotm (struct tm *result, char *input)
{
  char *p1=input, *p2;
  struct tm dtemp;
  time_t ttemp;
  int index;

  memset (&dtemp, '\0', sizeof (struct tm));
  dtemp.tm_isdst = -1;

  dtemp.tm_year = atoi (p1) - 1900;
  if ((p2 = strchr (p1, ' ')) == NULL || p1 == p2)
    return NULL;
  while (isspace (*p2)) p2++;

  for (index=1; mnames[index]; index++)
    if (_FP_strnicmp (p2, mnames[index], 3) == 0)
      break;
  if (mnames[index] == NULL)
    return NULL;
  dtemp.tm_mon = index - 1;

  while (*p2 && !isdigit (*p2)) p2++;
  if (*p2 == '\0') return NULL;
  dtemp.tm_mday = atoi (p2);

  if ((p1 = strchr (p2, ' ')) != NULL) {
    dtemp.tm_hour = atoi (++p1);

    if ((p2 = strpbrk (p1, " :")) != NULL) {
      dtemp.tm_min = atoi (++p2);
      if ((p1 = strpbrk (p2, " :")) != NULL)
	dtemp.tm_sec = atoi (++p1);
    }
  }
  if ((ttemp = mktime (&dtemp)) == (time_t)-1)
    return NULL;

  memcpy (result, &dtemp, sizeof (struct tm));
  return result;
}

struct tm *ctodate (struct tm *result, char *string)
{
  char *p1, *p2;
  time_t now;
  int index;

  now = time(NULL);
  memcpy (result, localtime (&now), sizeof (struct tm));

  result->tm_sec  = 0;
  result->tm_min  = 0;
  result->tm_hour = 0;
  result->tm_isdst=-1;

  if ((p1 = string) == NULL) return NULL;
  while (*p1 && !isalpha (*p1) && !isdigit (*p1) && *p1!='-') p1++;
  if (!isalpha (*p1) && !isdigit (*p1) && *p1!='-') return NULL;

  if (isalpha (*p1)) {
    for (index=1; mnames[index]; index++)
      if (_FP_strnicmp (p1, mnames[index], 3) == 0)
	break;
    if (mnames[index] == NULL)
      return NULL;
    result->tm_mon  = index - 1;
    while (*p1 && !isdigit (*p1)) p1++;
    if (*p1 != '\0') {
      result->tm_mday = atoi (p1);
      while (isdigit (*p1)) p1++;
      while (*p1 && !isdigit (*p1)) p1++;
      if (isdigit (*p1)) {
	p2 = p1;
	while (isdigit (*p2)) p2++;
	if (*p2 != ':') {
	  result->tm_year = (atoi (p1) > 1900) ? (atoi(p1)-1900) : atoi(p1);
	  p1 = p2;
	}
      }
    }
  }
  else if (atoi(p1)<=0) {
    now += atoi (p1) * 86400;

    memcpy (result, localtime (&now), sizeof (struct tm));

    result->tm_sec  = 0;
    result->tm_min  = 0;
    result->tm_hour = 0;

    while (*p1)
      p1++;
  }
  else {
    p2 = p1;
    while (isdigit (*p2)) p2++;
    if (*p2 == ':')
      goto ctdtime;
    if (*p2 == '/' || *p2 == '-')
      result->tm_mon  = atoi (p1) - 1;
    else
      result->tm_mday = atoi (p1);
    if (*p2 != '\0') {
      p1 = p2;
      while (*p1 && !isdigit (*p1)) p1++;
      if (*p1 != '\0') {
	if (*p2 == '/' || *p2 == '-')
	  result->tm_mday = atoi (p1);
	else
	  result->tm_mon  = atoi (p1) - 1;
	while (*p1 && isdigit (*p1)) p1++;
	if (*p1 == '/' || *p1 == '.') {
	  p1++;
	  if (isdigit (*p1))
	    result->tm_year = (atoi (p1) > 1900) ? (atoi(p1)-1900):(atoi(p1));
	  while (isdigit (*p1))
	    p1++;
	}
      }
    }
  }
  while (*p1 && !isdigit (*p1)) p1++;

 ctdtime:
  if (isdigit (*p1)) {
    result->tm_hour = atoi (p1);
    while (*p1 && *p1!=':') p1++;
    if (*p1 == ':') {
      result->tm_min = atoi (++p1);
      while (*p1 && *p1!=':') p1++;
      if (*p1 == ':')
	result->tm_sec = atoi (++p1);
    }
  }

  return result;
}

int dtdatecomp (dtree *d1, dtree *d2)
{
  int index, preres=0, level;
  char *p1, *p2;

  if (d1 == NULL || d2 == NULL)
    return 0;

  while (d1->depth > d2->depth) d1 = d1->parent;
  while (d1->depth < d2->depth) d2 = d2->parent;

  level = d1->depth;

  while (level >= 0 && d1 && d2) {
    if (level == 2) {
      p1 = d1->name; while (*p1 && !isdigit (*p1)) p1++;
      p2 = d2->name; while (*p2 && !isdigit (*p2)) p2++;
      if (atoi(p1)-atoi(p2) != 0)
	preres = atoi (p1) - atoi (p2);
    }
    else if (level == 1) {
      for (index=0; mnames[index]; index++) {
	if (strcmp (d1->name, mnames[index]) == 0) {
	  if (strcmp (d2->name, mnames[index]) == 0)
	    break;
	  preres = -1;
	  break;
	}
	else if (strcmp (d2->name, mnames[index]) == 0) {
	  preres =  1;
	  break;
	}
      }
    }
    else {
      if (atoi(d1->name) - atoi(d2->name) != 0)
	preres = atoi (d1->name) - atoi (d2->name);
    }
    d1 = d1->parent;
    d2 = d2->parent;
    level = level-1;
  }
  return preres;
}
     
/*
 * Tree functions
 */

static host *addhost_r (report *rep, char *hostname, host *depth, host *above, long size, int lev, int maxlev)
{
  host *iter=depth, *last=NULL, *new;
  char *level=strrchr (hostname, '.');

  if (level == NULL)
    level = hostname;
  else
    level++;

  while (iter) {
    if (strcmp (level, iter->name) <= 0)
      break;
    last = iter;
    iter = iter->NEXT;
  }

  if (iter && strcmp (level, iter->name) == 0) {
    iter->size += size >> 9;
    iter->req  += 1;

    if (level == hostname)
      return iter;
    else if (lev+1 == maxlev)
      return iter;
    else {
      *--level = '\0';
      last     = addhost_r (rep, hostname, iter->child, iter, size, lev+1, maxlev);
      *level   = '.';

      return last;
    }
  }
  if ((new = (host *) malloc (sizeof (host))) == NULL)
    return NULL;

  heapused += sizeof (host) + strlen (level) + 1;

  memset (new, '\0', sizeof (host));

  new->name   = _FP_strdup (level);
  new->req    = 1;
  new->size   = size >> 9;
  new->depth  = lev;
  new->parent = above;
  new->isleaf = (level != hostname) ? 0 : 1;
  new->NEXT   = iter;

  if (lev+1 == maxlev)
    iter = new;
  else if (level != hostname) {
    *--level = '\0';
    iter     = addhost_r (rep, hostname, NULL, new, size, lev+1, maxlev);
    *level++ = '.';
  }

  if (last == NULL) {
    if (above == NULL)
      rep->hlist = new;
    else
      above->child = new;
  }
  else
    last->NEXT = new;

  return ((level != hostname) ? iter : new);
}

static file *addfile_r (report *rep, char *filename, file *depth, file *above, long size, int lev, int maxlev)
{
  file *iter=depth, *last=NULL, *new;
  char *level=strchr(filename, '/');
  int flag;
  
  if (level == NULL) {
    if (*filename == '\0')
      filename = welcome;

    flag   = 1;
  }
  else {
    *level = '\0';
    flag   = 0;
  }

  while (iter) {
    if (strcmp (filename, iter->name) <= 0)
      break;
    last = iter;
    iter = iter->NEXT;
  }

  if (iter && strcmp (filename, iter->name) == 0) {
    iter->size += size >> 9;
    iter->req  += 1;

    if (flag == 0 && lev+1!=maxlev) {
      *level = '/';
      last   = addfile_r (rep, level+1, iter->child, iter, size, lev+1, maxlev);
    }
    else
      last = iter;

    return last;
  }
  if ((new = (file *) malloc (sizeof (file))) == NULL)
    return NULL;

  heapused += sizeof (file) + strlen (filename) + 1;

  memset (new, '\0', sizeof (file));

  new->name   = _FP_strdup (filename);
  new->req    = 1;
  new->size   = size >> 9;
  new->depth  = lev;
  new->parent = above;
  new->isleaf = (flag==0) ? 0 : 1;
  new->NEXT   = iter;

  if (flag==0 && lev+1!=maxlev) {
    *level = '/';
    iter   = addfile_r (rep, level+1, NULL, new, size, lev+1, maxlev);
  }
  else
    iter = new;

  if (last == NULL) {
    if (above == NULL)
      rep->flist = new;
    else
      above->child = new;
  }
  else
    last->NEXT = new;

  return ((flag==0 && lev+1!=maxlev) ? iter : new);
}

static dtree *adddate_r (report *rep, char *date, dtree *depth, dtree *above, long size, int level, int maxlev)
{
  char *sepa=strpbrk(date, " :"), *dp1, *dp2, schar;
  dtree *iter=depth, *last=NULL, *new;
  static char string[20];
  int val, cval, flag;

  while (isspace (*date))
    date++;

  if (level==2 && !isdigit (*date)) {
    if ((dp1 = sepa) == NULL)
      return NULL;
    sepa = strchr (++dp1, ' ');
  }
  else
    dp1 = date;

  if (sepa) {
    schar = *sepa;
    *sepa = '\0';
  }

  if (level==1) {
    for (val=0, flag=0; mnames[flag]; flag++)
      if (strcmp (dp1, mnames[flag]) == 0) {
	val = flag;
	break;
      }
  }
  else
    val = atoi (dp1);

  while (iter) {
    if (level==2 && !isdigit (*(iter->name))) {
      if ((dp2 = strchr (iter->name, ' ')) == NULL)
	return NULL;
      dp2++;
    }
    else
      dp2 = iter->name;

    if (level==1) {
      for (flag=0, cval=0; mnames[flag]; flag++)
	if (strcmp (dp2, mnames[flag]) == 0) {
	  cval = flag;
	  break;
	}
    }
    else
      cval = atoi (dp2);

    if (val <= cval)
      break;
    last = iter;
    iter = iter->NEXT;
  }

  if (iter && val==cval) {
    iter->size += size >> 9;
    iter->req  += 1;

    if (sepa && level+1!=maxlev)
      last = adddate_r (rep, sepa+1, iter->child, iter, size, level+1, maxlev);
    else
      last = iter;
    
    return last;
  }
  
  if ((new = (dtree *) malloc (sizeof (dtree))) == NULL)
    return NULL;

  memset (new, '\0', sizeof (dtree));

  if (level==2)
    sprintf (string, "%s %02d", above->name, val);
  else
    strcpy  (string, dp1);

  heapused += sizeof (dtree) + strlen (string) + 1;

  new->req    = 1;
  new->size   = size >> 9;
  new->name   = _FP_strdup (string);
  new->depth  = level;
  new->parent = above;
  new->isleaf = (sepa) ? 0 : 1;
  new->NEXT   = iter;

  if (sepa && level+1!=maxlev)
    iter = adddate_r (rep, sepa+1, NULL, new, size, level+1, maxlev);

  if (last == NULL) {
    if (above == NULL)
      rep->dlist = new;
    else
      above->child = new;
  }
  else
    last->NEXT = new;

  if (sepa)
    *sepa = schar;

  return (sepa && level+1!=maxlev) ? iter : new;
}

static entries *addentries (entries *old, entry *e)
{
  entries *new;

  if ((new = (entries *) malloc (sizeof (entries))) == NULL)
    return old;

  heapused += sizeof (entries);

  new->eptr = e;
  new->NEXT = old;

  return new;
}

int addentry (report *rep, char *hostname, char *filename,
	      char *date, long size,
	      int hostmax, int filemax, int datemax, int keepentries)
{
  entry *iter=rep->etail, *new;
  dtree *ndate;

  if (keepentries == 0) { /* don't memorize entries */
    if (hostmax != 0)
      (void) addhost (rep, hostname, size, hostmax);
    if (filemax != 0)
      (void) addfile (rep, filename, size, filemax);

    ndate = adddate (rep, date, size, (datemax>2 || datemax==-1)?datemax:3);

    if (ndate && (rep->mindate==NULL || dtdatecomp (ndate,rep->mindate)<0))
      rep->mindate = ndate;
    if (ndate && (rep->maxdate==NULL || dtdatecomp (ndate,rep->maxdate)>0))
      rep->maxdate = ndate;

    return 1;
  }

  if ((new = (entry *) malloc (sizeof (entry))) == NULL)
    return 0;

  heapused += sizeof (entry);

  if (hostmax == 0)
    new->ehost = NULL;
  else
    new->ehost = addhost (rep, hostname, size, hostmax);

  if (filemax == 0)
    new->efile = NULL;
  else
    new->efile = addfile (rep, filename, size, filemax);

  new->edate = adddate (rep, date, size, (datemax>2 || datemax==-1)?datemax:3);
  new->esize = size >> 9;
  new->NEXT  = NULL;

  if (new->edate&&(rep->mindate==NULL||dtdatecomp(new->edate,rep->mindate)<0))
    rep->mindate = new->edate;
  if (new->edate&&(rep->maxdate==NULL||dtdatecomp(new->edate,rep->maxdate)>0))
    rep->maxdate = new->edate;

  if (iter == NULL)
    rep->elist = new;
  else
    rep->etail->NEXT = new;

  rep->etail        = new;

  if (new->ehost)
    new->ehost->files = addentries (new->ehost->files, new);
  if (new->efile)
    new->efile->files = addentries (new->efile->files, new);
  if (new->edate)
    new->edate->files = addentries (new->edate->files, new);

  return 1;
}

char *getfqhost (host *hptr, char *buf)
{
  sprintf (buf, "%s%c", hptr->name, (hptr->parent) ? '.' : '\0');

  if (hptr->parent)
    getfqhost (hptr->parent, buf+strlen(buf));

  return buf;
}

char *getfqfile (file *fptr, char *buf)
{
  if (fptr->parent) {
    getfqfile (fptr->parent, buf);
    strcat    (buf, "/");
  }
  else
    *buf = '\0';

  strcat (buf, fptr->name);

  return buf;
}

char *getfqdate (dtree *dptr, char *buf)
{
  char *ptr;

  switch (dptr->depth) {
  case 0: /* year level */
    sprintf (buf, "%s", dptr->name);
    break;
  case 1: /* month level */
    sprintf (buf, "%s %s", dptr->parent->name, dptr->name);
    break;
  case 2: /* day level */
    ptr = dptr->name;
    while (*ptr && !isdigit (*ptr))
      ptr++;

    sprintf (buf, "%s %s %s",
	     dptr->parent->parent->name,
	     dptr->parent->name,
	     ptr);
    break;
  case 3: /* hour level */
    ptr = dptr->parent->name;
    while (*ptr && !isdigit (*ptr))
      ptr++;
    sprintf (buf, "%s %s %s %s:00",
	     dptr->parent->parent->parent->name,
	     dptr->parent->parent->name,
	     ptr,
	     dptr->name);
    break;
  case 4: /* min level */
    ptr = dptr->parent->parent->name;
    while (*ptr && !isdigit (*ptr))
      ptr++;
    sprintf (buf, "%s %s %s %s:%s",
	     dptr->parent->parent->parent->parent->name,
	     dptr->parent->parent->parent->name,
	     ptr,
	     dptr->parent->name,
	     dptr->name);
    break;
  case 5: /* sec level */
    ptr = dptr->parent->parent->parent->name;
    while (*ptr && !isdigit (*ptr))
      ptr++;
    sprintf (buf, "%s %s %s %s:%s:%s",
	     dptr->parent->parent->parent->parent->parent->name,
	     dptr->parent->parent->parent->parent->name,
	     ptr,
	     dptr->parent->parent->name,
	     dptr->parent->name,
	     dptr->name);
    break;
  default:
    *buf = '\0';
    break;
  }
  return buf;
}

/*
 * char *getfqdate (dtree *dptr, char *buf)
 * {
 *   if (dptr->parent) {
 *     getfqdate (dptr->parent, buf);
 *     strcat    (buf, " ");
 *   }
 *   else
 *     *buf = '\0';
 * 
 *   strcat (buf, dptr->name);
 * 
 *   return buf;
 * }
 */

int getentry (entry *ent, char *hostname, char *filename, char *date, int *size)
{
  if (ent==NULL || ent->ehost==NULL || ent->efile==NULL)
    return 0;

  getfqhost (ent->ehost, hostname);
  getfqfile (ent->efile, filename);
  getfqdate (ent->edate, date);
  *size = ent->esize << 9;

  return 1;
}

host *locatehost_r (host *iter, char *hostname)
{
  char *level=strrchr (hostname, '.');
  host *result;

  if (level == NULL)
    level = hostname;
  else
    level++;

  while (iter) {
    if (strcmp (level, iter->name) <= 0)
      break;
    iter = iter->NEXT;
  }

  if (iter==NULL || strcmp (level, iter->name) != 0)
    return NULL;

  if (level == hostname)
    result = iter;
  else {
    *--level = '\0';
    result   = locatehost_r (iter->child, hostname);
    *level   = '.';
  }

  return result;
}

file *locatefile_r (file *iter, char *filename)
{
  char *level=strchr(filename, '/');
  file *result;
  int flag;
  
  if (level == NULL) {
    if (*filename == '\0')
      filename = welcome;

    flag   = 1;
  }
  else {
    *level = '\0';
    flag   = 0;
  }

  while (iter) {
    if (strcmp (filename, iter->name) <= 0)
      break;
    iter = iter->NEXT;
  }

  if (iter==NULL || strcmp (filename, iter->name) != 0) {
    if (flag == 0)
      *level = '/';

    return NULL;
  }

  if (flag == 0 && *(level+1)) {
    *level = '/';
    result = locatefile_r (iter->child, level+1);
  }
  else
    result = iter;

  return result;
}

dtree *locatedate_r (dtree *iter, char *date)
{
  char *sepa=strchr(date, ' '), *dp1, *dp2;
  int val, cval, flag;
  dtree *result;

  if (iter->depth==2 && !isdigit (*date)) {
    if ((dp1 = strchr (date, ' ')) == NULL)
      return NULL;
    sepa = strchr (++dp1, ' ');
  }
  else
    dp1 = date;

  if (sepa)
    *sepa = '\0';

  if (iter->depth==1) {
    for (val=0, flag=0; mnames[flag]; flag++)
      if (strcmp (dp1, mnames[flag]) == 0) {
	val = flag;
	break;
      }
  }
  else
    val = atoi (dp1);

  while (iter) {
    if (iter->depth==2 && !isdigit (*(iter->name))) {
      if ((dp2 = strchr (iter->name, ' ')) == NULL)
	return NULL;
      dp2++;
    }
    else
      dp2 = iter->name;

    if (iter->depth==1) {
      for (flag=0, cval=0; mnames[flag]; flag++)
	if (strcmp (dp2, mnames[flag]) == 0) {
	  cval = flag;
	  break;
	}
    }
    else
      cval = atoi (dp2);

    if (val <= cval)
      break;

    iter = iter->NEXT;
  }

  if (iter && val==cval) {
    if (sepa)
      result = locatedate_r (iter->child, sepa+1);
    else
      result = iter;
  }
  else
    result = NULL;
  
  if (sepa)
    *sepa = ' ';

  return result;
}

static void *sortbyrs_r (void *list, int type, int sstrat)
{
  void **child, *iter=list, *new=NULL, *insert;
  void **nlptr, *last, *next, **nptr;
  long siz1, siz2;
  int req1, req2;

  if (list == NULL)
    return NULL;

  while (iter) {
    insert = new;
    last   = NULL;

    switch (type) {
    case HOSTTYPE:
      child= (void **) &((host*)iter)->child;
      req1 =            ((host*)iter)->req;
      siz1 =            ((host*)iter)->size;
      nptr = (void **) &((host*)iter)->NEXT;
      break;
    case FILETYPE:
      child= (void **) &((file*)iter)->child;
      req1 =            ((file*)iter)->req;
      siz1 =            ((file*)iter)->size;
      nptr = (void **) &((file*)iter)->NEXT;
      break;
    case DATETYPE:
      child= (void **) &((dtree*)iter)->child;
      req1 =            ((dtree*)iter)->req;
      siz1 =            ((dtree*)iter)->size;
      nptr = (void **) &((dtree*)iter)->NEXT;
      break;
    }
    *child = sortbyrs_r (*child, type, sstrat);
    next   = *nptr;

    while (insert) {
      switch (type) {
      case HOSTTYPE: req2=((host*)insert)->req; siz2=((host*)insert)->size; break;
      case FILETYPE: req2=((file*)insert)->req; siz2=((file*)insert)->size; break;
      case DATETYPE: req2=((dtree*)insert)->req;siz2=((dtree*)insert)->size;break;
      }
      if ((sstrat==SORTREQ) ? (req1>req2) : (siz1>siz2))
	break;

      last = insert;

      switch (type) {
      case HOSTTYPE: nlptr = (void**) &((host*)last)->NEXT; break;
      case FILETYPE: nlptr = (void**) &((file*)last)->NEXT; break;
      case DATETYPE: nlptr = (void**) &((dtree*)last)->NEXT;break;
      }
      insert = *nlptr;
    }
    if (last==NULL) {
      *nptr = new;
      new   = iter;
      iter  = next;
      continue;
    }
    *nlptr = iter;
    *nptr  = insert;
    iter   = next;
  }
  return new;
}

static entry *sortentry (entry *list, int sstrat)
{
  entry *newlist=NULL, *lastins=NULL, *insert, *last, *next;
  int cval;

  if (sstrat != SORTALPHA && sstrat != SORTSIZE)
    return list;

  while (list) {
    if (lastins) {
      switch (sstrat) {
      case SORTALPHA:
	cval = dtdatecomp (list->edate, lastins->edate);
	break;
      case SORTSIZE:
	cval = (list->esize < lastins->esize) ? -1 : 1;
	break;
      }
      if (cval >= 0) {
	insert = lastins->NEXT;
	last   = lastins;
      }
      else {
	insert = newlist;
	last   = NULL;
      }
    }
    else {
      insert = newlist;
      last   = NULL;
    }
    while (insert) {
      switch (sstrat) {
      case SORTALPHA:
	cval = dtdatecomp (list->edate, insert->edate);
	break;
      case SORTSIZE:
	cval = (list->esize < insert->esize) ? -1 : 1;
	break;
      }
      if (cval <= 0)
	break;
      last   = insert;
      insert = insert->NEXT;
    }
    next = list->NEXT;

    if (last == NULL) {
      list->NEXT = newlist;
      newlist    = list;
    }
    else {
      list->NEXT = insert;
      last->NEXT = list;
    }
    lastins = list;
    list    = next;
  }
  return newlist;
}

report *sortbyrs (report *rep, int type, int sstrat)
{
  entry *iter;

  if (rep==NULL ||
      (type==ENTRYTYPE && sstrat!=SORTALPHA && sstrat!=SORTSIZE) ||
      (type!=ENTRYTYPE && sstrat!=SORTREQ   && sstrat!=SORTSIZE && type!=0))
    return rep;

  switch (type) {
  case HOSTTYPE:
    if (sstrat==SORTREQ || sstrat==SORTSIZE)
      rep->hlist = (host *) sortbyrs_r (rep->hlist, HOSTTYPE, sstrat);
    break;
  case FILETYPE:
    if (sstrat==SORTREQ || sstrat==SORTSIZE)
      rep->flist = (file *) sortbyrs_r (rep->flist, FILETYPE, sstrat);
    break;
  case DATETYPE:
    if (sstrat==SORTREQ || sstrat==SORTSIZE)
      rep->dlist = (dtree*) sortbyrs_r (rep->dlist, DATETYPE, sstrat);
    break;
  case ENTRYTYPE:
    if ((iter = rep->elist = (entry *) sortentry  (rep->elist, sstrat))!=NULL)
      while (iter->NEXT)
	iter = iter->NEXT;
    rep->etail = iter;
    break;
  default:
    if (sstrat==SORTREQ || sstrat==SORTSIZE) {
      rep->hlist = (host *) sortbyrs_r (rep->hlist, HOSTTYPE, sstrat);
      rep->flist = (file *) sortbyrs_r (rep->flist, FILETYPE, sstrat);
      rep->dlist = (dtree*) sortbyrs_r (rep->dlist, DATETYPE, sstrat);
    }
    if ((iter = rep->elist = (entry *) sortentry  (rep->elist, sstrat))!=NULL)
      while (iter->NEXT)
	iter = iter->NEXT;
    rep->etail = iter;
    break;
  }
  return rep;
}

void *getmaxreq (void *what, int type, int level)
{
  void *result=NULL, *next, *iter=what, *child;
  int maxsofar=0, val;

  while (iter) {
    switch (type) {
    case HOSTTYPE: 
      next = ((host *)iter)->NEXT;
      val  = ((host *)iter)->req;
      if (level) {
	if ((child=getmaxreq (((host*)iter)->child, HOSTTYPE, level-1))!=NULL)
	  if (((host*)child)->req > maxsofar) {
	    result   = child;
	    maxsofar = ((host*)child)->req;
	  }
      }
      break;
    case FILETYPE: 
      next = ((file *)iter)->NEXT;
      val  = ((file *)iter)->req;
      if (level) {
	if ((child=getmaxreq (((file*)iter)->child, FILETYPE, level-1))!=NULL)
	  if (((file*)child)->req > maxsofar) {
	    result   = child;
	    maxsofar = ((file*)child)->req;
	  }
      }
      break;
    case DATETYPE: 
      next = ((dtree*)iter)->NEXT;
      val  = ((dtree*)iter)->req;
      if (level) {
	if ((child=getmaxreq (((dtree*)iter)->child, HOSTTYPE, level-1))!=NULL)
	  if (((dtree*)child)->req > maxsofar) {
	    result   = child;
	    maxsofar = ((dtree*)child)->req;
	  }
      }
      break;
    default: return NULL;
    }
    if (level == 0) {
      if (val > maxsofar) {
	result   = iter;
	maxsofar = val;
      }
    }
    iter = next;
  }
  return result;
}

int countleafs (void *what, int type, int level)
{
  void *next, *iter=what;
  int sofar=0;

  if (iter == NULL)
    return 0;

  while (iter) {
    if (level == 0)
      sofar++;

    switch (type) {
    case HOSTTYPE: 
      next = ((host *)iter)->NEXT;
      if (level) {
	if (((host*)iter)->child == NULL)
	  sofar++;
	else
	  sofar += countleafs (((host *)iter)->child, HOSTTYPE, level-1);
      }
      break;
    case FILETYPE: 
      next = ((file *)iter)->NEXT;
      if (level)  {
	if (((file*)iter)->child == NULL)
	  sofar++;
	else
	  sofar += countleafs (((file *)iter)->child, FILETYPE, level-1);
      }
      break;
    case DATETYPE: 
      next = ((dtree*)iter)->NEXT;
      if (level) {
	if (((dtree*)iter)->child == NULL)
	  sofar++;
	else
	  sofar += countleafs (((dtree*)iter)->child, HOSTTYPE, level-1);
      }
      break;
    default:
      return 0;
    }
    iter = next;
  }
  return sofar;
}

static report *newreport_r (report *rep, void *what, int type)
{
  static char filename[256], hostname[256], date[256];
  void *childs, *iter=what, *next;
  entries *eiter;
  int size;

  while (iter) {
    switch (type) {
    case HOSTTYPE:
      eiter  = ((host *) iter)->files;
      childs = ((host *) iter)->child;
      next   = ((host *) iter)->NEXT;
      break;
    case FILETYPE:
      eiter  = ((file *) iter)->files;
      childs = ((file *) iter)->child;
      next   = ((file *) iter)->NEXT;
      break;
    case DATETYPE:
      eiter  = ((dtree *) iter)->files;
      childs = ((dtree *) iter)->child;
      next   = ((dtree *) iter)->NEXT;
      break;
    }

    while (eiter) {
      if (getentry (eiter->eptr, hostname, filename, date, &size))
	addentry (rep, hostname, filename, date, size, -1, -1, -1, 1);
      eiter = eiter->NEXT;
    }

    if (childs)
      rep = newreport_r (rep, childs, type);

    iter = next;
  }
  return rep;
}

report *newreport (void *what, int type)
{
  static char filename[256], hostname[256], date[256];
  entries *eiter;
  void *childs;
  report *new;
  int size;

  if ((new = (report *) malloc (sizeof (report))) == NULL)
    return NULL;

  heapused += sizeof (report);

  memset (new, '\0', sizeof (report));

  if (what == NULL)
    return new;

  switch (type) {
  case HOSTTYPE:
    eiter  = ((host *) what)->files;
    childs = ((host *) what)->child;
    break;
  case FILETYPE:
    eiter  = ((file *) what)->files;
    childs = ((file *) what)->child;
    break;
  case DATETYPE:
    eiter  = ((dtree *) what)->files;
    childs = ((dtree *) what)->child;
    break;
  default:
    return new;
  }

  while (eiter) {
    if (getentry (eiter->eptr, hostname, filename, date, &size))
      addentry (new, hostname, filename, date, size, -1, -1, -1, 1);
    eiter = eiter->NEXT;
  }

  if (childs)
    new = newreport_r (new, childs, type);

  return new;
}

static void killentries (entries *liste)
{
  entries *iter = liste, *last;

  while (iter) {
    last = iter;
    iter = iter->NEXT;

    heapused -= sizeof (entries);

    free (last);
  }
}

static void killhost (host *tree)
{
  host *iter = tree, *last;

  while (iter) {
    killentries (iter->files);
    killhost    (iter->child);

    last = iter;
    iter = iter->NEXT;

    heapused -= sizeof (host) + strlen (last->name) + 1;

    free (last->name);
    free (last);
  }
}

static void killfile (file *tree)
{
  file *iter = tree, *last;

  while (iter) {
    killentries (iter->files);
    killfile    (iter->child);

    last = iter;
    iter = iter->NEXT;

    heapused -= sizeof (file) + strlen (last->name) + 1;

    free (last->name);
    free (last);
  }
}

static void killdtree (dtree *tree)
{
  dtree *iter = tree, *last;

  while (iter) {
    killentries (iter->files);
    killdtree   (iter->child);

    last = iter;
    iter = iter->NEXT;

    heapused -= sizeof (dtree) + strlen (last->name) + 1;

    free (last->name);
    free (last);
  }
}

void killreport (report *rep)
{
  entry *iter, *last;

  if (rep == NULL)
    return;
  iter = rep->elist;

  while (iter) {
    last = iter;
    iter = iter->NEXT;

    heapused -= sizeof (entry);

    free (last);
  }
  killhost (rep->hlist);
  killfile (rep->flist);
  killdtree(rep->dlist);

  heapused -= sizeof (report);

  free (rep);
}
