/*
 * This file is part of W3Olista, the statistics gathering program for
 * the WWW.                  This sourcecode is (c) Frank Pilhofer '95
 *
 * It is  distributed without warranty. The complete license terms for
 * this software  are included in  the file 'License.html'  in the doc
 * subdirectory. If  your copy came  without this file, you should get
 * it on the Web at    http://www.uni-frankfurt.de/~fp/Tools/W3Olista/
 * or contact the author,               fp@informatik.uni-frankfurt.de
 */


/*
 * Definition of structures. Welcome to pointermania
 */

struct _entry;
struct tm;

/*
 * Report / structure types
 */

#define NOTYPE		0
#define HOSTTYPE	1
#define FILETYPE	2
#define DATETYPE	3
#define ENTRYTYPE	4

/*
 * Sort types
 */

#define SORTALPHA	0
#define SORTREQ		1
#define SORTSIZE	2

typedef struct _entries {
  struct _entry   *eptr;
  struct _entries *NEXT;
} entries;

/*
 * Because the long int limit of 2 gigabytes is reached surprisingly
 * fast, all sizes are now recorded >>9 (512 byte chunks). This'll allow
 * for reporting of up to 1 terabyte of transmitted data.
 */

typedef struct _host {
  char    *name;
  int      req, depth;
  int      isleaf;
  long     size;
  entries *files;

  struct _host *parent;
  struct _host *child;
  struct _host *NEXT;
} host;

typedef struct _file {
  char    *name;
  int      req, depth;
  int      isleaf;
  long     size;
  entries *files;

  struct _file *parent;
  struct _file *child;
  struct _file *NEXT;
} file;

/*
 * Date format used here:
 *     "1995" (Year)
 *     "Feb"  (English month, 3 chars, first capital)
 *     "6"    (Day of Month, 01-31, may again be preceded by month name)
 *     "10"   (Hour of Day,  01-24)
 * separated by exactly one space
 */

typedef struct _dtree {
  char    *name;
  int      req, depth;
  int      isleaf;
  long     size;
  entries *files;

  struct _dtree *parent;
  struct _dtree *child;
  struct _dtree *NEXT;
} dtree;

typedef struct _entry {
  host   *ehost;
  file   *efile;
  dtree  *edate;
  long    esize;

  struct _entry *NEXT;
} entry;

typedef struct _report {
  entry  *elist, *etail;
  host   *hlist;
  file   *flist;
  dtree  *dlist;
  dtree  *mindate;
  dtree  *maxdate;
} report;

extern int heapused;
extern char *mnames[], *mlnames[];

char *		tmtodtree	(struct tm *);
struct tm *	dtreetotm	(struct tm *, char *);
struct tm *	ctodate		(struct tm *, char *);
int		dtdatecomp	(dtree *, dtree *);

int		addentry	(report *, char *, char *, char *, long, int, int, int, int);
char *		getfqhost	(host *, char *);
char *		getfqfile	(file *, char *);
char *		getfqdate	(dtree *, char *);
int		getentry	(entry *, char *, char *, char *, int *);
host *		locatehost_r	(host *, char *);
file *		locatefile_r	(file *, char *);
dtree *		locatedate_r	(dtree *, char *);
void *		getmaxreq	(void *, int, int);
int		countleafs	(void *, int, int);
report *	newreport	(void *, int);
void		killreport	(report *);
report *	sortbyrs	(report *, int, int);

#define locatehost(r,s)		locatehost_r(r->hlist,s)
#define locatefile(r,s)		locatefile_r(r->flist,s)
#define locatedate(r,s)		locatedate_r(r->dlist,s)
